#include <stdio.h>
#include <stdlib.h>
#include <pc.h>
#include <dpmi.h>
#include <go32.h>
#include <dos.h>

#ifndef USENEAR
#include <sys/farptr.h>
#else
#include <sys/nearptr.h>
#include <crt0.h>
int _crt0_startup_flags = _CRT0_FLAG_NEARPTR | _CRT0_FLAG_NONMOVE_SBRK;
#endif

#define BYTE unsigned char

typedef struct {
  unsigned ModeAttributes;
  unsigned granularity,startseg,farfunc;
  short bscanline;
  short XResolution;
  short YResolution;
  short charpixels;
  unsigned bogus1,bogus2,bogus3,bogus4;
  unsigned PhysBasePtr;
  char bogus[228];
} ModeInfoBlock;

unsigned int ADDR;
int width,height;

ModeInfoBlock *get_mode_info(int mode)
{
	static ModeInfoBlock info;
	__dpmi_regs r;

	/* Use the transfer buffer to store the results of VBE call */
	r.x.ax = 0x4F01;
	r.x.cx = mode;
	r.x.es = __tb / 16;
	r.x.di = 0;
	__dpmi_int(0x10, &r);
	if(r.h.ah) return 0;
	dosmemget(__tb, sizeof(ModeInfoBlock), &info);
	return &info;
}

void set_palette(int entry,int r,int g,int b)
{
   outportb(0x3c8,(BYTE)entry);
   outportb(0x3c9,(BYTE)r);
   outportb(0x3c9,(BYTE)g);
   outportb(0x3c9,(BYTE)b);
}

void init_graphics(void)
{
	__dpmi_meminfo info;
	__dpmi_regs reg;
	ModeInfoBlock *mb;

	mb=get_mode_info(0x101);

	if (!mb) {
	  printf("Get VESA mode info failed.\n");
	  exit(1);
	}
	if (!(mb->ModeAttributes & 0x80)) {
	  printf("Linear frame buffer not supported for VESA mode.\n");
	  exit(1);
	}

	width=mb->XResolution;
	height=mb->YResolution;
	info.size=width*height;
	info.address=mb->PhysBasePtr;
	if(__dpmi_physical_address_mapping(&info) == -1) {
	  printf("Physical mapping of address 0x%x failed!\n",mb->PhysBasePtr);
	  exit(2);
	}
	ADDR=info.address;		/* Updated by above call */
	printf("Vesa mode 0x101: %d x %d, linear frame: 0x%x\n",width,height,ADDR);
	delay(1500);

	reg.x.ax=0x4f02;
	reg.x.bx=0x4101;		/* mode plus linear enable bit */
	__dpmi_int(0x10,&reg);		/* set the mode */
	if(reg.h.al != 0x4f || reg.h.ah) {
	  printf("Mode set failed!\n");
	  exit(3);
	}
/*	for (l=0;l<64;l++) {
		BYTE d,r,g,b;
		int l,r2;
		d=(BYTE) l;
		r=(BYTE) l;
		r2=(l-32)*2;
		if (r2<0) r2=0;
		g=(BYTE)r2;
		set_palette(d,r,r,r);
	}*/
}

void close_graphics(void)
{
	__dpmi_regs reg;
	reg.x.ax = 0x0003;
	__dpmi_int(0x10,&reg);
}

void main(void)
{
   int i;

#ifndef USENEAR
   int my_video_ds;
   printf("Test VBE 2.0 linear buffer support with far pointers:\n");
   init_graphics();
   /* Note: Video mode probably doesn't support text writes at this point */
   my_video_ds = __dpmi_allocate_ldt_descriptors(1);
   __dpmi_set_segment_base_address(my_video_ds,ADDR);
   ADDR = 0;	/* basis now zero from selector */
   __dpmi_set_segment_limit(my_video_ds,(width*height)|0xfff);

   _farpokeb(my_video_ds,0,15);	/* put a pixel at top corner */

   for(i=0;i<width;i++) {
     _farnspokeb(ADDR+i,50);
     _farnspokeb(ADDR+i+width*(height-1),51);
     _farnspokeb(ADDR+i+width*((height-1)*i/width),52);
     _farnspokeb(ADDR+width-1-i+width*((height-1)*i/width),53);
   }
#else
   char *video;
   printf("Test VBE 2.0 linear buffer support with near pointers:\n");
   init_graphics();
   /* Note: Video mode probably doesn't support text writes at this point */
   video = (char *)(ADDR + __djgpp_conventional_base);
   video[0] = 15;			/* put a pixel at top corner */

   malloc(1024*1024);			/* See if sbrk() moves us */
   for(i=0;i<width;i++) {
     video[i] = 50;
     video[i+width*(height-1)] = 51;
     video[i+width*((height-1)*i/width)] = 52;
     video[width-1-i+width*((height-1)*i/width)] = 53;
   }
#endif
   getkey();
   close_graphics();
}
